"""
Compute the proximity operator of some indicator functions.

Author: Mbaye DIONGUE (Jun 2022)
"""

from .BoxConstraint import *
from .EpiAbsValue import *
from .EpiDistance import *
from .EpiExp import *
from .EpiHingeLoss import *
from .EpiL2Norm import *
from .EpiLinfNorm import *
from .EpiLog import *
from .EpiMax import *
from .EpiPower import *
from .EpiSquaredHingeLoss import *
from .EpiSupport import *
from .ExtendedSimplex import *
from .Hyperplane import *
from .Hyperslab import *
from .L1_2Ball import *
from .L1Ball import *
from .L2Ball import *
from .LinfBall import *
from .LogProjection import *
from .MonotoneCone import *
from .PositiveL2Ball import *
from .Simplex import *


__all__ = [
    "BoxConstraint",
    "EpiAbsValue",
    "EpiDistance",
    "EpiExp",
    "EpiHingeLoss",
    "EpiL2Norm",
    "EpiLinfNorm",
    "EpiLog",
    "EpiMax",
    "EpiPower",
    "EpiSquaredHingeLoss",
    "EpiSupport",
    "ExtendedSimplex",
    "Hyperplane",
    "Hyperslab",
    "L1_2Ball",
    "L1Ball",
    "L2Ball",
    "LinfBall",
    "LogProjection",
    "MonotoneCone",
    "PositiveL2Ball",
    "Simplex"
]
